# coding=utf8
""" Output

Output methods
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__version__		= "1.0.0"
__email__		= "chris@ouroboroscoding.com"
__created__		= "2022-03-03"

# Python imports
import sys

# Pip imports
from termcolor import colored

def error(msg):
	"""Error

	Print bold red text to stderr

	Arguments:
		msg (str): The message to print

	Returns:
		None
	"""
	sys.stderr.write(
		colored(
			msg,
			color='red',
			attrs=['bold']
		)
	)

def color(color_, msg):
	"""Color

	Prints bold messages in a specific color

	Arguments:

		arg (str): The message to print

	Returns:
		None
	"""
	sys.stdout.write(
		colored(
			msg,
			color=color_,
			attrs=['bold']
		)
	)

def verbose(msg):
	"""Verbose

	Print bold white text to stdout

	Arguments
		msg (str): The message to print

	Returns:
		None
	"""
	sys.stdout.write(
		colored(
			msg,
			color='white',
			attrs=['bold']
		)
	)



# coding=utf8
""" Py Live Dev

Goes through the passed script and keeps track of imported modules so that the
script can be re-started on any updates
"""

from __future__ import print_function

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__version__		= "1.0.0"
__email__		= "chris@ouroboroscoding.com"
__created__		= "2021-06-05"

# Python imports
from time import sleep

# Local imports
from .app import App
from .output import color

def main(conf):
	"""Main

	Primary entry point into the script

	Arguments:
		conf (dict): The configuration for the apps to run

	Returns:
		bool
	"""

	color('magenta', 'Starting PyLiveDev\n')

	# Init the list of apps
	lApps = []

	# Check for a defaults section
	oDefaults = conf.pop('__defaults__', None)

	# Go through each app in the config
	for name in conf:

		# If we have defaults, override them with the specific app config
		if oDefaults:
			oConf = {**oDefaults, **conf[name]}

		# Else, just use the conf as is
		else:
			oConf = conf[name]

		# Init the type and list of files associated
		oApp = App(name, **oConf)

		# Start the app
		if oApp.start():
			lApps.append(oApp)

	# If we have no apps
	if not lApps:
		return False

	# Loop endlessly until we're asked to shut down
	try:
		while True:
			sleep(1)
	except KeyboardInterrupt:
		for o in lApps:
			o.stop()
			del o

	color('magenta', '\nGoodbye\n')

	# Return OK
	return True



# coding=utf8
""" Imports

Handles looking for and loading imports recursively
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__version__		= "1.0.0"
__email__		= "chris@ouroboroscoding.com"
__created__		= "2021-06-05"

# Python imports
import ast
import os

# Local imports
from . import output

def exists(name, folder=None):
	"""Exists

	Checks if a given import name exists as a local file. If it does, a string
	with the filename is returned, else False

	Arguments:
		name (str): The name of the import to check

	Returns
		bool|str
	"""

	# If we have a local folder and it doesn't end in /
	if folder and folder[-1:] != '/':
		folder = '%s/' % folder

	# Init the parts
	lParts = []

	# If we have a name
	if name:

		# Init the . count
		iDots = 0
		while name[iDots] == '.':
			iDots += 1

		# Add .. folders for each dot (we ignore the first one)
		for i in range(1, iDots):
			lParts.append('..')

		# Split the name by .
		lParts.extend(name.lstrip('.').split('.'))

	# While we have a count to the parts
	while len(lParts):

		# Generate the project filename
		sFile = '%s.py' % '/'.join(lParts)

		# Does it exist locally from the parent?
		if folder:

			# Generate the local filename
			sLocalFile = '%s%s' % (folder, sFile)

			# Does the file exist in it
			if os.path.exists(sLocalFile):
				return sLocalFile

		# Does it exist off the project
		if os.path.exists(sFile):
			return sFile

		# Does it exist as an __init__?
		sInit = '%s/__init__.py' % '/'.join(lParts)
		if os.path.exists(sInit):
			return sInit

		# Still doesn't exist? Let's take a piece off the parts and try again
		lParts = lParts[0:-1]

	# Nothing found, return False
	return False

def find(file, file_list):
	"""Find

	Looks through the file for import statements and if the files are local,
	adds them to the list

	Arguments:
		file (str): The name of the file to open and parse looking for imports
		file_list (list): The unique list of existing files as well as where new
							files will be added

	Returns:
		None
	"""

	# Open the file
	with open(file) as oF:

		# Get the abstract syntax tree for the file
		try:
			oAST = ast.parse(oF.read(), file)

			# Go through each node in the tree
			for oNode in ast.iter_child_nodes(oAST):

				# If the instance is an import
				if isinstance(oNode, ast.Import):

					# Go through each name found
					for oName in oNode.names:

						# Look for a file
						mFile = exists(oName.name, os.path.dirname(file))

						# If the file exists
						if mFile:

							# If it doesn't exist already in the list
							if mFile not in file_list:

								# Add it
								file_list.append(mFile)

								# And recurse
								find(mFile, file_list)

				# If the instance is a from
				elif isinstance(oNode, ast.ImportFrom):

					# If there's no module
					if not oNode.module:
						mFile = exists('__init__', os.path.dirname(file))
					else:
						mFile = exists(oNode.module, os.path.dirname(file))

					# If the file exists
					if mFile:

						# If it doesn't exist already in the list
						if mFile not in file_list:

							# Add it
							file_list.append(mFile)

							# And recurse
							find(mFile, file_list)

					# Go through each name found
					for oName in oNode.names:

						# Look for a file
						mFile = exists(
							'%s.%s' % (oNode.module is not None and oNode.module or '', oName.name),
							os.path.dirname(file)
						)

						# If the file exists
						if mFile:

							# If it doesn't exist already in the list
							if mFile not in file_list:

								# Add it
								file_list.append(mFile)

								# And recurse
								find(mFile, file_list)

		# Catch syntax errors from broken code
		except SyntaxError as e:
			output.error('Syntax Error parsing "%s" at line %d, column %d\n' % (
				file,
				e.args[1][1],
				e.args[1][2]
			))



# coding=utf8
""" App

Handles the class passed around with all data related to a single script
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__version__		= "1.0.0"
__email__		= "chris@ouroboroscoding.com"
__created__		= "2021-06-05"

# Python imports
import os
import subprocess
import sys

# Pip imports
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

# Local imports
from . import imports, output

class App(FileSystemEventHandler):
	"""App

	class to hold all relevant data for running a single application

	Extends
		object
	"""

	def __init__(self, name, command, \
					additional_files=None, arguments=None, mode=None, \
					python=None, unbuffered=True, verbose=False):
		"""Constructor

		Handles instantiating the App instance

		Arguments:
			name (str): The name of the app
			command (str): The command to run to start the script
			additional_files (str[]): Additional files to watch that aren't python imports
			arguments (str[]): Additional arguments to pass to the script
			mode (str): The mode of script, 'script' or 'module'
			python (str): The full path to the python to use
			unbuffered (bool): If true, script will run in unbuffered mode for stdout/stderr
			verbose (bool): If true, additional data will be displayed when running script

		Returns:
			App
		"""

		# Init configurable members
		self._name = name
		self._command = command
		self._additional_files = additional_files
		self._arguments = arguments
		self._mode = mode and mode or 'script'
		self._python = python or sys.executable
		self._unbuffered = unbuffered
		self._verbose = verbose

		# Init internal members
		self._files = []
		self._fresh_line = True

		# Create a new observer
		self._observer = Observer()
		self._observer.start()

		# Generate the arguments to run the script/module
		self._generate_args()

	def __del__(self):
		"""Deconstructor

		Cleans up after the instance

		Returns
			None
		"""

		# Stop the observer
		if self._observer:
			self._observer.stop()
			self._observer.join()
			del self._observer

	def _generate_args(self):
		"""Generate Args

		Private method to generate the full list of arguments based on the data
		associated with the app

		Returns:
			None
		"""

		# Init the list with the python executable
		self._args = [self._python]

		# If we are running unbuffered
		if self._unbuffered:
			self._args.append('-u')

		# If we are running a module
		if self._mode == 'module':
			self._args.append('-m')

		# Add the script/module
		self._args.append(self._command)

		# If there's additional arguments to the script
		if self._arguments and isinstance(self._arguments, list):
			self._args.extend(self._arguments)

		# If verbose, display list of arguments
		if self._verbose:
			output.verbose('The following args were generated: %s\n' % str(self._args))

	def dispatch(self, event):
		"""Dispatch

		Called whenever there's an event on any of the watched files

		Returns:
			None
		"""

		# If it's a modified file and it's in our list
		if not event.is_directory and \
			event.event_type == 'modified' and \
			event.src_path in self._files:

			# If verbose mode is on, notify of a file change
			if self._verbose:
				output.verbose('%s has been modified\n' % event.src_path)

			# Stop the app
			self.stop()

			# Start the app
			self.start()

	def join(self):
		"""Join

		Passes join message along to observer

		Returns:
			None
		"""
		self._observer.join()

	def start(self):
		"""Start

		Starts the app by first parsing the imports and adding observers for
		them, then running the actual process.

		Returns
			None
		"""

		# If verbose mode is on
		if self._verbose:
			output.verbose('Starting %s\n' % self._name)

		# Clear the files
		self._files = []

		# If it's a module
		if self._mode == 'module':

			if self._verbose:
				output.verbose('\tobserving module\n')

			# Convert . to /
			sFile = self._command.replace('.', '/')

			# Does the file exist as is?
			if os.path.exists('%s.py' % sFile):
				self._files.append('%s.py' % sFile)

			# Else, look for special python module file(s)
			else:

				# Check for an __init__.py file
				if os.path.exists('%s/__init__.py' % sFile):
					self._files.append('%s/__init__.py' % sFile)

				# Check for a __main__.py file
				if os.path.exists('%s/__main__.py' % sFile):
					self._files.append('%s/__main__.py' % sFile)

		# Else, it's a script
		else:

			if self._verbose:
				output.verbose('\tobserving script\n')

			# Check for the command as is
			if os.path.exists(self._command):
				self._files.append(self._command)

		# If we have no files
		if not self._files:

			# Print error
			output.error('Can not find anything to load for %s\n' % self._name)

			# Return error
			return False

		# Go through each found file
		for sFile in list(self._files):

			# Look for more files within it
			imports.find(sFile, self._files)

		# If verbose mode is on
		if self._verbose:
			output.verbose('\tthe following imports were found:\n')
			for s in self._files:
				output.verbose('\t\t%s\n' % s)

		# Add additional files if the exist
		if self._additional_files:

			# Verbose
			if self._verbose:
				output.verbose('\tthe following additional files were found:\n')

			# Go through each file and make sure it exists
			for sFile in self._additional_files:
				if os.path.exists(sFile):
					self._files.append(sFile)
					if self._verbose:
						output.verbose('\t\t%s\n' % sFile)
				else:
					output.error('"%s" does not exist\n' % sFile)

		# For each file, add to the obverver
		for s in self._files:
			self._observer.schedule(self, s)

		# Create the subprocess
		try:
			if self._verbose:
				output.verbose('\tcreating subprocess...')

			self._process = subprocess.Popen(
				self._args,
				bufsize=0,
				cwd=os.getcwd(),
				env=os.environ,
				stdout=sys.stdout,
				stderr=sys.stderr
			)

			if self._verbose:
				output.verbose(' done\n')

		except OSError as e:
			output.error('%s: invalid process\n%s\n' % (self._name, str(e.args)))
			return False

		except ValueError as e:
			output.error('%s: invalid arguments\n%s\n' % (self._name, str(e.args)))
			return False

		# Return OK
		return True

	def stop(self):
		"""Stop

		Stops the app, first by removing all observers, then by stopping the
		actual process

		Returns:
			None
		"""

		# If verbose mode is on
		if self._verbose:
			output.verbose('Stopping %s\n' % self._name)

		# Stop watching all associated files
		if self._verbose:
			output.verbose('\tstop observing files...')
		self._observer.unschedule_all()
		if self._verbose:
			output.verbose(' done\n')

		# If we have a process
		if self._process:

			# Send a signal to the process to terminate
			if self._verbose:
				output.verbose('\tterminating process...')
			self._process.terminate()

			# Wait for the process to terminate
			try:
				self._process.wait(10)

			# If it won't shut down, kill it
			except subprocess.TimeoutExpired:
				if self._verbose:
					output.verbose('\nProcessing not terminating, attempting to kill...')
				self._process.kill()

			if self._verbose:
				output.verbose(' done\n')

		# Delete the process
		del self._process



from setuptools import setup

with open('README.md', 'r') as oF:
	long_description=oF.read()

setup(
	name='pylivedev',
	version='1.0.0',
	description='PyLiveDev is used to keep track of files associated with your script so it can be re-started if any file is updated.',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://ouroboroscoding.com/pylivedev',
	project_urls={
		'Source': 'https://github.com/ouroboroscoding/pylivedev',
		'Tracker': 'https://github.com/ouroboroscoding/pylivedev/issues'
	},
	keywords=['python','live', 'development'],
	author='Chris Nasr - OuroborosCoding',
	author_email='chris@ouroboroscoding.com',
	license='Apache-2.0',
	packages=['pylivedev'],
	install_requires=[
		'termcolor>=1.1.0',
		'watchdog>=2.1.2'
	],
	entry_points={
		'console_scripts': ['pylivedev=pylivedev.__main__:cli']
	},
	zip_safe=True
)
